from collections import namedtuple

import requests

TRACKING_URI = 'https://ssl.google-analytics.com/collect'


def _request(data, extra_headers):
    return requests.post(TRACKING_URI, data=data, headers=extra_headers)


def report(tracking_id, client_id, requestable, extra_info=None,
           extra_headers=None):
    """Actually report measurements to Google Analytics."""
    return [_request(data, extra_headers)
            for data, extra_headers in payloads(
            tracking_id, client_id, requestable, extra_info, extra_headers)]


def payloads(tracking_id, client_id, requestable, extra_info=None,
             extra_headers=None):
    """Get data and headers of API requests for Google Analytics.

    Generates a sequence of (data, headers) pairs. Both `data` and `headers`
    are dicts.
    """
    extra_payload = {
        'v': '1',
        'tid': tracking_id,
        'cid': client_id,
        'aip': '1'}
    if extra_info:
        for payload in extra_info:
            extra_payload.update(payload)

    for request_payload in requestable:
        final_payload = dict(request_payload)
        final_payload.update(extra_payload)
        yield final_payload, extra_headers


class Requestable(object):

    def get_payload(self):
        raise NotImplementedError()

    def __iter__(self):
        yield self.get_payload()


class SystemInfo(Requestable, namedtuple('SystemInfo', 'language')):

    def __new__(cls, language=None):
        return super(SystemInfo, cls).__new__(cls, language)

    def get_payload(self):
        payload = {}
        if self.language:
            payload['ul'] = self.language
        return payload


class PageView(
        Requestable,
        namedtuple('PageView',
                   'path host_name location title referrer')):

    def __new__(cls, path=None, host_name=None, location=None, title=None,
                referrer=None):
        return super(PageView, cls).__new__(cls, path, host_name, location,
                                            title, referrer)

    def get_payload(self):
        payload = {'t': 'pageview'}
        if self.location:
            payload['dl'] = self.location
        if self.host_name:
            payload['dh'] = self.host_name
        if self.path:
            payload['dp'] = self.path
        if self.title:
            payload['dt'] = self.title
        if self.referrer:
            payload['dr'] = self.referrer
        return payload


class Event(Requestable, namedtuple('Event', 'category action label value')):

    def __new__(cls, category, action, label=None, value=None):
        return super(Event, cls).__new__(cls, category, action, label, value)

    def get_payload(self):
        payload = {
            't': 'event',
            'ec': self.category,
            'ea': self.action}
        if self.label:
            payload['el'] = self.label
        if self.value:
            payload['ev'] = str(int(self.value))
        return payload


class Transaction(
        Requestable,
        namedtuple('Transaction',
                   'transaction_id items revenue shipping affiliation')):

    def __new__(cls, transaction_id, items, revenue=None, shipping=None,
                affiliation=None):
        if not items:
            raise ValueError('You need to specify at least one item')
        return super(Transaction, cls).__new__(
            cls, transaction_id, items, revenue, shipping, affiliation)

    def get_total(self):
        if self.revenue:
            return self.revenue
        prices = [i.get_subtotal() for i in self.items]
        total = sum(prices[1:], prices[0])
        if self.shipping:
            total += self.shipping
        return total

    def get_payload(self):
        payload = {
            't': 'transaction',
            'ti': self.transaction_id}
        if self.affiliation:
            payload['ta'] = self.affiliation
        total = self.get_total()
        payload['tr'] = str(total.gross)
        payload['tt'] = str(total.tax)
        payload['cu'] = total.currency
        if self.shipping:
            payload['ts'] = str(self.shipping.gross)
        return payload

    def __iter__(self):
        yield self.get_payload()
        for i in self.items:
            yield i.get_payload_for_transaction(self.transaction_id)


class Item(namedtuple('Item', 'name unit_price quantity item_id category')):

    def __new__(cls, name, unit_price, quantity=None, item_id=None,
                category=None):
        return super(Item, cls).__new__(cls, name, unit_price, quantity,
                                        item_id, category)

    def get_subtotal(self):
        if self.quantity:
            return self.unit_price * self.quantity
        return self.unit_price

    def get_payload_for_transaction(self, transaction_id):
        payload = {
            't': 'item',
            'ti': transaction_id,
            'in': self.name}
        payload['ip'] = str(self.unit_price.gross)
        payload['cu'] = self.unit_price.currency
        if self.quantity:
            payload['iq'] = str(int(self.quantity))
        if self.item_id:
            payload['ic'] = self.item_id
        if self.category:
            payload['iv'] = self.category
        return payload



#! /usr/bin/env python
from setuptools import setup

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.5',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

setup(name='google-measurement-protocol',
      author='Mirumee Software',
      author_email='hello@mirumee.com',
      description=('A Python implementation of'
                   ' Google Analytics Measurement Protocol'),
      license='BSD',
      version='0.1.3',
      packages=['google_measurement_protocol'],
      install_required=['requests>=2.0,<3.0a0'],
      test_suite='google_measurement_protocol.tests',
      tests_require=['httmock>=1.0,<1.1a0', 'prices>=0.5,<0.6a0'],
      classifiers=CLASSIFIERS,
      platforms=['any'])



from unittest import TestCase
try:
    from urllib.parse import parse_qs
except ImportError:
    from urlparse import parse_qs

from httmock import response, urlmatch, with_httmock
from prices import Price

from . import (Event, Item, PageView, report, SystemInfo, Requestable,
               Transaction, payloads)


class MockRequestable(Requestable):

    def get_payload(self):
        return {'t': 'mock'}


@urlmatch(netloc=r'ssl\.google-analytics\.com', path='/collect')
def ga_mock(url, request):
    qs = parse_qs(request.body)
    return response(200, qs)


class ReportTest(TestCase):

    @with_httmock(ga_mock)
    def test_report(self):
        mr = MockRequestable()
        (response,) = report('UA-123456-78', 'CID', mr)
        data = response.json()
        self.assertEqual(data['cid'], ['CID'])
        self.assertEqual(data['tid'], ['UA-123456-78'])
        self.assertEqual(data['t'], ['mock'])

    @with_httmock(ga_mock)
    def test_extra_info(self):
        empty_info = SystemInfo()
        self.assertEqual(empty_info.get_payload(), {})
        mr = MockRequestable()
        info = SystemInfo(language='en-gb')
        (response,) = report('UA-123456-78', 'CID', mr, extra_info=info)
        data = response.json()
        self.assertEqual(data['ul'], ['en-gb'])


class PageViewTest(TestCase):

    def test_by_path(self):
        pv1 = PageView('/my-page/')
        self.assertEqual(pv1.get_payload(),
                         {'t': 'pageview', 'dp': '/my-page/'})
        pv2 = PageView('/my-page/', host_name='example.com')
        self.assertEqual(
            pv2.get_payload(),
            {'t': 'pageview', 'dp': '/my-page/', 'dh': 'example.com'})

    def test_by_location(self):
        view = PageView(location='http://example.com/my-page/')
        self.assertEqual(
            view.get_payload(),
            {'t': 'pageview', 'dl': 'http://example.com/my-page/'})

    def test_optional_params(self):
        view = PageView('/', title='title', referrer='referrer')
        self.assertEqual(
            view.get_payload(),
            {'t': 'pageview', 'dp': '/', 'dr': 'referrer', 'dt': 'title'})


class EventTest(TestCase):

    def test_required_params(self):
        evt = Event('category', 'action')
        self.assertEqual(evt.get_payload(),
                         {'t': 'event', 'ec': 'category', 'ea': 'action'})

    def test_optional_params(self):
        evt = Event('category', 'action', label='label', value=7)
        self.assertEqual(
            evt.get_payload(),
            {'t': 'event', 'ec': 'category', 'ea': 'action', 'el': 'label',
             'ev': '7'})


class ItemTest(TestCase):

    def test_required_params(self):
        item = Item('item-01', Price(10, currency='USD'))
        self.assertEqual(
            item.get_payload_for_transaction('trans-01'),
            {'t': 'item', 'in': 'item-01', 'cu': 'USD', 'ip': '10',
             'ti': 'trans-01'})

    def test_quantity(self):
        item = Item('item-01', Price(10, currency='USD'), quantity=2)
        self.assertEqual(
            item.get_payload_for_transaction('trans-01'),
            {'t': 'item', 'in': 'item-01', 'cu': 'USD', 'ip': '10',
             'iq': '2', 'ti': 'trans-01'})
        self.assertEqual(item.get_subtotal(), Price(20, currency='USD'))

    def test_optional_params(self):
        item = Item('item-01', Price(10, currency='USD'), item_id='it01',
                    category='cat')
        self.assertEqual(
            item.get_payload_for_transaction('trans-01'),
            {'t': 'item', 'in': 'item-01', 'cu': 'USD', 'ip': '10',
             'ic': 'it01', 'iv': 'cat', 'ti': 'trans-01'})


class TransactionTest(TestCase):

    def test_no_items(self):
        self.assertRaises(ValueError, lambda: Transaction('trans-01', []))

    def test_required_params(self):
        items = [Item('item-01', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items)
        self.assertEqual(
            trans.get_payload(),
            {'t': 'transaction', 'ti': 'trans-01', 'cu': 'USD', 'tr': '10',
             'tt': '0'})

    def test_revenue_override(self):
        items = [Item('item-01', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items,
                            revenue=Price(net=40, gross=50, currency='USD'))
        self.assertEqual(
            trans.get_payload(),
            {'t': 'transaction', 'ti': 'trans-01', 'cu': 'USD', 'tr': '50',
             'tt': '10'})

    def test_shipping(self):
        items = [Item('item-01', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items,
                            shipping=Price(100, currency='USD'))
        self.assertEqual(
            trans.get_payload(),
            {'t': 'transaction', 'ti': 'trans-01', 'cu': 'USD', 'tr': '110',
             'ts': '100', 'tt': '0'})

    def test_affiliation(self):
        items = [Item('item-01', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items, affiliation='loyalty')
        self.assertEqual(
            trans.get_payload(),
            {'t': 'transaction', 'ti': 'trans-01', 'cu': 'USD', 'tr': '10',
             'tt': '0', 'ta': 'loyalty'})

    def test_iter(self):
        items = [Item('item-01', Price(10, currency='USD')),
                 Item('item-02', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items)
        trans_payloads = list(trans)
        self.assertEqual(len(trans_payloads), 3)


class PayloadsTest(TestCase):

    def test_payloads(self):
        items = [Item('item-01', Price(10, currency='USD')),
                 Item('item-02', Price(10, currency='USD'))]
        trans = Transaction('trans-01', items)
        trans_payloads = list(payloads(
            'tracking-id',
            'client-id',
            trans,
            SystemInfo(language='en-gb'),
            {'extra-header-key': 'extra-head-value'}))

        self.assertEqual(len(trans_payloads), 3)
        for data, headers in trans_payloads:
            self.assertEqual(data['tid'], 'tracking-id')
            self.assertEqual(data['cid'], 'client-id')
            self.assertEqual(data['ul'], 'en-gb')
            self.assertTrue(headers['extra-header-key'], 'extra-header-value')
